# Custom exception classes

class TSMBase(Exception):
	"""Base class for the Telstra Smart Modem"""

class TSMAuthError(TSMBase):
	"""Errors relating to authentication"""

class TSMModemError(TSMBase):
	"""Errors relating to the modem not working correctly"""



# Class to represent the devices connected to or seen by the modem.
# This class is returned from Modem.getclients() and can't be used by itself.

import re
import bs4

# Compiled regular expressions:
re_mac = re.compile(r"(?:[0-9a-f]{2}[:]){5}[0-9a-f]{2}")
re_ipv4 = re.compile(r"(?:[0-9]{1,3}[.]){3}[0-9]{1,3}")
re_ipv6_full = re.compile(r"(?:[0-9a-f]{1,4}:){7}[0-9a-f]{1,4}")


class Devices():

	def __init__(self, html):
		self.devices = self._parse(html)

	# Parse the html table from Modem.getclients()
	def _parse(self, html):

		headers = ['online', 'hostname', 'ip', 'mac', 'connection', 'eth-port']

		def parseOnlineStatus(td: bs4.element.Tag) -> bool:
			status = td.div['class'][1]
			if status == "green":
				return True
			else:
				return False
		
		def parseEthPort(possiblePort: str):
			if possiblePort is not None:
				return int(possiblePort)

		def parseIPV4(ips: str):
			if ips is not None:
				match = re_ipv4.search(ips)
				if match is not None:
					return match.group()

		def parseIPV6(ips: str):
			if ips is None:
				return []
			else:
				return re_ipv6_full.findall(ips)


		html_table = html.tbody.find_all('tr')

		devices = []

		for row in html_table:
			cols = row.find_all('td')
			device = {}
			for index, item in enumerate(cols):
				header = headers[index]

				if index == 0:
					device[header] = parseOnlineStatus(item)
				elif index == 2:
					device['ipv4'] = parseIPV4(item.string)
					device['ipv6'] = parseIPV6(item.string)
				elif index == 5:
					device[header] = parseEthPort(item.string)
				else:
					device[header] = item.string

			devices.append(device)

		return devices

	# Validate a supplied MAC address and convert it to the correct format.
	def validMAC(self, mac: str) -> str:
		valid_mac = mac.lower()
		valid_mac = valid_mac.replace("-", ":", 5)

		if re_mac.fullmatch(valid_mac) is None:
			raise ValueError(f"Invalid MAC address: '{mac}'")

		return valid_mac

	# Get a specific device's details from a MAC address.
	def getDevice(self, mac: str) -> dict:
		valid_mac = self.validMAC(mac)
		for device in self.devices:
			if device['mac'] == valid_mac:
				return device

	# Check if a device is connected to the modem by MAC address.
	def isOnline(self, mac: str) -> bool:
		device = self.getDevice(mac)
		if device is None:
			return False
		else:
			return device['online']



# Non standard implementation of the SRP protocol that Telstra/Technicolor
# uses for the smart modem gen 2 and possibly other Technicolor routers as well. Since
# I couldn't get any Python SRP libraries working correctly, this implementation is
# reverse engineered from srp-min.js (sent from the modem). It only functions as a client/user.

import hashlib
import random

# Convert Int to hex string
def toHex(number: int) -> str:
	hexString = format(number, 'x')
	if 1 == len(hexString) % 2:
		hexString = '0' + hexString
	return hexString

# SHA256 hash helper
def hash(value: bytes, outformat: str = 'HEX'):
	h = hashlib.sha256()
	h.update(value)
	if outformat == 'BIN':
		return h.digest()
	elif outformat == 'INT':
		return int.from_bytes(h.digest(), byteorder='big')
	else:
		return h.hexdigest()

# Calculate the 'u' variable in SRP
def calculate_u(A: str, B: str) -> int:
	A_bytes = A.to_bytes(256, 'big', signed=False)
	B_bytes = B.to_bytes(256, 'big', signed=False)
	hashed = hash(A_bytes + B_bytes, outformat='INT')
	return hashed

# Get Random integer for the 'a' client variable in SRP
def getRandomA() -> int:
	return random.SystemRandom().getrandbits(2048) % N


# Shared SRP variables:
g = 2
N = '''
ac6bdb41324a9a9bf166de5e1389582faf72b6651987ee07fc3192943db56050a37329cbb4a099ed8193e0757767a13dd52312ab4b03310d
cd7f48a9da04fd50e8083969edb767b0cf6095179a163ab3661a05fbd5faaae82918a9962f0b93b855f97993ec975eeaa80d740adbf4ff74
7359d041d5c33ea71d281e446b14773bca97b43a23fb801676bd207a436c6481f1d2b9078717461a5b9d32e688f87748544523b524b0d57d
5ea77a2775d2ecfa032cfbdbf52fb3786160279004e57ae6af874e7303ce53299ccc041c7bc308d82a5698f3a8d0c38271ae35f8e9dbfbb6
94b5c803d89f7ae435de236d525f54759b65e372fcd68ef20fa7111f9e4aff73
'''
N = N.replace(" ", "").replace("\n", "")
N = int(N, 16)

# Magic numbers:
magic_C = '05b9e8ef059c6b32ea59fc1d322d37f04aa30bae5aa9003b8321e21ddb04e300'
magic_C = int(magic_C, 16)
# This one stays in hex format. Likely some kind of shared secret.
magic_u = '4a76a9a2402bdd18123389b72ebbda50a30f65aedb90d7273130edea4b29cc4c'

# SRP User class for authenticating with the Telstra Smart Modem.
class User:

	def __init__(self, username: str, password: str):
		self.username = username
		self.password = password

	# Return the calculated 'A' variable in SRP to send to the server.
	def start_authentication(self) -> str:
		self.a = getRandomA()

		A = 0
		while A == 0:
			A = pow(g, self.a, N)

		self.A = A
		self.A_hex = toHex(A)
		return self.A_hex

	# Take the 's' and 'B' SRP variables from the server and calculate 
	# the 'M' SRP variable to send back to the server.
	def process_challenge(self, s_hex: str, B_hex: str) -> str:
		B = int(B_hex, 16)
		u = calculate_u(self.A, B)

		user_pass = hash(f"{self.username}:{self.password}".encode())
		n = hash(bytes.fromhex(s_hex + user_pass), 'INT')

		mya = (magic_C * pow(g, n, N)) % N
		myb = (self.a + ((u * n) % N)) % N
		mye = pow(((B - mya) % N), myb, N)

		mye_bytes = mye.to_bytes(256, byteorder='big')

		_B = hash(mye_bytes)
		_e = hash(self.username.encode())
		M = hash(bytes.fromhex(magic_u + _e + s_hex + self.A_hex + B_hex + _B))
		return M



import setuptools

with open("README.md") as f:
	long_description = f.read()

setuptools.setup(
	name="telstra-smart-modem",
	version="1.0.0",
	author="hatdz",
	description="Library to retrieve information from a Telstra Smart Modem.",
	license="MIT",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/hatdz/telstra-smart-modem",
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	python_requires='>=3.6',
	install_requires=[
		'requests>=2.22,<3',
		'beautifulsoup4>=4.8.2,<5'
	]
)



# Library for retrieving information from a Telstra Smart Modem.
# Based on scraping the web interface which takes a very long time to respond.
# It is the only way of retrieving infortmation from the modem without hacking it.

import ipaddress

import requests
import bs4

import telstra_smart_modem.srp as tsm_srp
import telstra_smart_modem.devices as tsm_devices
import telstra_smart_modem.exceptions as tsm_errors

HTTP_TIMEOUT = (3.05, 6)


class Modem():

	session = requests.Session()
	session.hooks = {
		'response': lambda r, *args, **kwargs: r.raise_for_status()
	}
	CSRFtoken = None

	def __init__(self, ip, username, password):
		host = ipaddress.IPv4Address(ip)
		self.base_url = f"http://{host}"
		self.username = username
		self.password = password

		self._authenticate()


	# Extract the CSRFtoken from html returned by one of the modem's web pages.
	def _extractCSRFtoken(self, html):
		soup = bs4.BeautifulSoup(html, 'html.parser')
		CSRFtoken = soup.head.find('meta', attrs={'name': 'CSRFtoken'})
		if CSRFtoken is not None:
			self.CSRFtoken = CSRFtoken['content']
		else:
			raise tsm_errors.TSMModemError("Expected CSRFtoken but didn't find one.")


	# Extract the CSRFtoken initially using html returned from the modem's index page.
	def _getCSRFtoken(self):
		index = self.session.get(self.base_url, timeout=HTTP_TIMEOUT)
		self._extractCSRFtoken(index.text)


	# Authenticate with the Modem using SRP.
	def _authenticate(self, html=None):
		
		def firstAuth(A: str):
			authenticate_first = self.session.post(
				f"{self.base_url}/authenticate",
				{
					"CSRFtoken": self.CSRFtoken,
					"I": self.username,
					"A": A
				},
				timeout=HTTP_TIMEOUT
			)
			first_response = authenticate_first.json()
			s = first_response['s']
			B = first_response['B']
			return s, B


		def secondAuth(M: str):
			authenticate_second = self.session.post(
				f"{self.base_url}/authenticate",
				{
					"CSRFtoken": self.CSRFtoken,
					"M": M
				},
				timeout=HTTP_TIMEOUT
			)
			second_response = authenticate_second.json()
			return second_response
		
		
		try:
			if html is not None:
				self._extractCSRFtoken(html)
			elif self.CSRFtoken is None:
				self._getCSRFtoken()

			srp = tsm_srp.User(self.username, self.password)
			A = srp.start_authentication()
			s, B = firstAuth(A)
			M = srp.process_challenge(s, B)
			second_response = secondAuth(M)
		
		except Exception as e:
			raise tsm_errors.TSMModemError(f"Error during authentication: {e}")

		# Check if password is correct
		recv_M = second_response.get('M')
		if recv_M is None:
			raise tsm_errors.TSMAuthError('Username and/or password is incorrect')


	# Logout of the modem. (For testing purposes)
	def _logout(self):
		try:
			logout = self.session.post(
				f"{self.base_url}/login.lp",
				{
					"do_signout": 1,
					"CSRFtoken": self.CSRFtoken
				},
				timeout=HTTP_TIMEOUT
			)
			_extractCSRFtoken(logout.text)
		except Exception as e:
			raise tsm_errors.TSMModemError(f"Error during logout: {e}")


	# Helper function to re-authenticate if timed-out.
	def _tryGet(self, modalFunction, exception, errorMessage):
		successful, data = modalFunction()

		if not successful:
			self._authenticate(data)
			successful, data = modalFunction()
			if not successful:
				raise exception(errorMessage)
		
		return data


	# Return a Devices object of all devices seen by the modem.
	def getDevices(self):
		
		def getDeviceModal():
			devices_response = self.session.get(
				f"{self.base_url}/modals/device-modal.lp",
				timeout=HTTP_TIMEOUT
			)
			soup = bs4.BeautifulSoup(devices_response.text, 'html.parser')
			html_table = soup.find('table', attrs={
				"id": "devices",
				"class": "table table-striped"
			})
			if html_table is None:
				return False, devices_response.text
			else:
				return True, html_table

		data = self._tryGet(
			getDeviceModal,
			tsm_errors.TSMModemError,
			"Failed to get clients from the modem."
		)
		return tsm_devices.Devices(data)


	# Return the status of the modem. (online, backup or offline)
	def getModemStatus(self):

		def getStatusModal():
			status_response = self.session.get(self.base_url, timeout=HTTP_TIMEOUT)
			soup = bs4.BeautifulSoup(status_response.text, 'html.parser')
			status = soup.find('div', attrs={
				"class": "message"
			})
			if status is None:
				return False, status_response.text
			else:
				return True, status

		def parseStatus(classname):
			if classname == "ok":
				return "online"
			elif classname == "backup":
				return "backup"
			elif classname == "error":
				return "offline"

		data = self._tryGet(getStatusModal, tsm_errors.TSMModemError, "Failed to get modem status.")

		img_class = data.find('img', attrs={
			"src": "img/status.png"
		})['class'][0]

		return parseStatus(img_class)
